/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * Released under GNU GPL v2. */

/* matcher_tlewis.c: a simple matcher implementation for hank */

#include "hank.h"
#include "event.h"
#include "packet.h"
#include "protocol_engine.h"
#include "protocol_lookup.h"
#include "support.h"
#include <libxml/debugXML.h>

#define IS_INV(test, val) (test ? !val : val)

static char rcsid[] = "$Id: matcher_tlewis.c,v 1.21 2001/12/08 00:04:48 tlewis Exp $";

extern struct protocol_lookup pls;
extern lt proto_table;

enum ETYPE {
	ETYPE_DECOMP,
	ETYPE_COL,
	ETYPE_AND,
	ETYPE_OR,
	ETYPE_PRIM,
	ETYPE_INVALID
};

/* The all-singing, all-dancing match element structure. */
struct melement {
	enum ETYPE disc;
	unsigned int id; /* local to parent */
	struct melement *parent;
	char not;
	struct melement *next;
	struct melement *kids;
	u_int32_t protocol;
	prim_id prid;
	event_id event;
};

/* helpers */
static struct melement* add_element(struct melement* parent, enum ETYPE type);
static struct melement *add_primitive(struct melement* parent, char *proto, char *section, char *field, char *op, char *arg, char not);
static int build_rule(struct melement *parent, xmlNodePtr x, event_id e, int inversion);
static void decompose(struct packet *p);

/* exports */
static event_id create_event(void);
static handler_id register_handler(event_id incoming, event_id outgoing, xmlNodePtr match);
static int unregister_handler(handler_id handler);
static void handle_event(struct packet *p);
static int set_handler_target(event_id incoming, handler_id handler, event_id target);

struct event_handler matcher = {
	{
		"event_handler",
		"matcher",
		EHMAJ,
		EHMIN,
		&matcher
	},
	&create_event,
	&register_handler,
	&unregister_handler,
	&set_handler_target,
	&handle_event
};

static struct melement *collections[1024];
static int collections_used=0;

#define STATIC_MODULE_NAME matcher_tlewis

#ifdef ENABLE_DYN_MODULES
struct smodule * paengine [2] = {
#else  /*ENABLE_DYN_MODULES*/
struct smodule * STATIC_MODULE_NAME [2] =  {
#endif /*ENABLE_DYN_MODULES*/
	&(matcher.s),
	NULL
};

#define PRINTF(x...) 

/* helper implementations */

static int
build_rule(struct melement *parent, xmlNodePtr x, event_id e, int inversion)
{
	struct melement *m=NULL;
	xmlNodePtr y;

	/* recursion rocks */
	x=get_real_xml_node(x);
	if(x&&!strcmp((const char *)x->name, "decomp")){
		m=add_element(parent, ETYPE_DECOMP);
		PRINTF("Decomp target added as %i.\n", m->id);

	} else if(x&&!strcmp((const char *)x->name, "comment")){
		/* skip */
	} else if(x&&!strcmp((const char *)x->name, "NOT")){
		PRINTF("INVERTED: ");
		for(y=x->children;y;y=y->next){
			if(strcmp((const char *)y->name, "text"))
				if(build_rule(parent, y, e, 1)!=0)
					return(-1);
				return(0);
		}
		return(-1);

	} else if(x&&!strcmp((const char *)x->name, "AND")){
		if((m=add_element(parent, ETYPE_AND))){
			for(y=x->children;y;y=y->next){
				if(strcmp((const char *)y->name, "text"))
					build_rule(m, y, 0, 0);
			}
		}

	} else if(x&&!strcmp((const char *)x->name, "OR")){
		if((m=add_element(parent, ETYPE_OR)))
			for(y=x->children;y;y=y->next){
				if(strcmp((const char *)y->name, "text"))
					build_rule(m, y, 0, 0);
			}

	} else if(x&&!strcmp((const char *)x->name, "PRIM")) {
		char *proto, *section, *field, *op, *arg=NULL;
		int not=0;

		proto=(char *)xmlGetProp(x, (unsigned const char *)"proto");
		section=(char *)xmlGetProp(x, (unsigned const char *)"section");
		field=(char *)xmlGetProp(x, (unsigned const char *)"field");
		op=(char *)xmlGetProp(x, (unsigned const char *)"op");
		if(x->children)
			arg=(char *)x->children->content;

		if(strcmp(op, "present") && !field){
			/* this operation is not "present" */
			fprintf(stderr, "Bogus rule: operation is \"%s\", but field is empty.\n", op);
			m=NULL;
		} else {
			m=add_primitive(parent, proto, section, field, op, arg, not);
			if(m==NULL)
				fprintf(stderr, "ERROR at %s:%i: bad primitive: %s: %s, %s, %s, %s\n",
					x->doc->URL, xmlGetLineNo(x),
					proto, section, field, op, arg);
		}


	} else {
		/* FIXME: log; error. */
		if(x)
			printf("Unrecognized node type: %s.  Error.\n", x->name);
		else
			printf("Empty node; error.\n");
		return(-1);
	}

	if(m==NULL){
		return(-1);
	}

	m->not=inversion;
	m->event=e;

	return(0);
}

/* determines if an element matches.  Recursively callable */
/* returns 0 on failure */ 
static int
eval_element(struct packet *p, struct melement *t)
{
	struct capsule *e, *f;
	struct melement *m;
	struct capsule *c;

	PRINTF("ENTERING: eval_element, melement is %p, type is %i.%i.\n", t, t->disc>>24, t->disc&0xffffff);
	fflush(stdout);
	
	if (t->disc == ETYPE_PRIM) {
		unsigned long itchy, scratchy;
		rdtscl(itchy);
		f=NULL;
		for(e=p->encaps;e;e=e->next){
			/* find encap */
			if(e->protocol==t->protocol){
				f=e;
				break;
			}
		}
		rdtscl(scratchy);
		PRINTF("Finding encap took %i cycles.\n", (scratchy-itchy));

		if(f==NULL){
			/* encap not found for protocol, normal */
			PRINTF("LEAVING: PRIM: no encap.\n");
			return(IS_INV(t->not, 0));
		}

		if(f->matches==NULL){
			PRINTF("LEAVING: No matches for protocol.\n");
			return(IS_INV(t->not, 0));
		}

		PRINTF("LEAVING: Result for protocol %i.%i, prid %i is (not=%i) %i.\n", t->protocol>>24, t->protocol&0xffffff, t->prid, t->not, f->matches[t->prid]);

		return(IS_INV(t->not, f->matches[t->prid]));
	} else if (t->disc  == ETYPE_AND) {
		for(m=t->kids;m;m=m->next){
			if(!eval_element(p, m)){
				PRINTF("LEAVING: AND: child false.\n");
				return(0);
			}
		}
		PRINTF("LEAVING: AND: true.\n");
		return(IS_INV(t->not, 1));
	} else if (t->disc == ETYPE_DECOMP) {
		unsigned long foo, bar, baz;
		rdtscl(foo);
		decompose(p);
		rdtscl(bar);
		for(c=p->encaps;c;c=c->next){
			if(c->pe){
				unsigned long coke, pepsi;
				rdtscl(coke);
				c->pe->find_matches(c, NULL);
				rdtscl(pepsi);
				printf("Finding matches for %s took %ul\n", c->pe->s.name, (pepsi-coke));
			} else {
				/*  FIXME:
				until pe's supplied for paengines, this
				will always happen for the first encap.
				printf("ERROR: pe not supplied in encap!\n");
				*/
			}
		}
		rdtscl(baz);
		printf("Decomp took %ul, matching took %ul cycles.\n", (bar-foo),
			(baz-bar));
		PRINTF("LEAVING: Decomp complete.\n");
		return(IS_INV(t->not, 1));
	} else if(t->disc == ETYPE_COL){
		/* FIXME: ERROR: can't eval a COL */
		PRINTF("Aiee!  Can't evaluate a collection.\n");
		return(0);
	} else if (t->disc == ETYPE_OR) {
		for(m=t->kids;m;m=m->next){
			if(eval_element(p, m)){
				PRINTF("LEAVING: OR: child true.\n");
				return(IS_INV(t->not, 1));
			}
		}
		PRINTF("LEAVING: OR: false.\n");
		return(IS_INV(t->not, 0));
	}

	printf("Unrecognized type, %i.%i, in melement at %p.\n", t->disc>>24, t->disc&0xffffff, t);
	return(0);

}

static struct melement*
add_element(struct melement* parent, enum ETYPE type)
{
	struct melement *e = NULL, *f;

	if ( (type != ETYPE_AND) && (type != ETYPE_OR) && \
			(type != ETYPE_COL) && (type != ETYPE_DECOMP) ){
		printf("Unrecognized element type.\n");
		return(NULL);
	}

/*
	if(parent)
		printf("Adding element to %i.\n", parent->id);
	else
		printf("Adding top-level element.\n");
*/

	if(!(e=malloc(sizeof(*e))))
		/* FIXME */
		exit(-1);

	e->disc=type;
	e->id=0;
	e->parent=parent;
	e->not=0;
	e->next=NULL;
	e->kids=NULL;
	e->protocol=0;
	e->prid=0;
	e->event=0;

	if(parent==NULL){
		if (type != ETYPE_COL) {
			printf("No parent, & type != COL.\n");
			free(e);
			return(NULL);
		}
	} else {
		if (type == ETYPE_COL) {
			printf("Parent, & type == COL.\n");
			free(e);
			return(NULL);
		}

		if(parent->kids){
			e->id=(parent->kids->id + 1);
			/* find last */
			f=parent->kids;
			while(f->next)
				f=f->next;
			PRINTF("Setting f->next to %p.\n", e);
			f->next=e;
		} else {
			PRINTF("Setting f->kids to %p.\n", e);
			parent->kids=e;
		}

	}

	return(e);
}

static struct melement*
add_primitive(struct melement* parent, char *proto, char *section, char *field, char *op, char *arg, char not)
{
	struct melement *e = NULL;
	struct protocol_engine *pe;

	if(!(e=malloc(sizeof(*e))))
		/* FIXME */
		exit(-1);

	e->disc=ETYPE_PRIM;
	e->id=0;
	e->parent=parent;
	e->not=not;
	e->next=NULL;
	e->kids=NULL;

	if((pe=pls.find_pe_name(proto_table, proto))){
		PRINTF("Adding primitive for protocol %s.\n", pe->s.name);
		e->protocol=pe->protocols[0];
	} else {
		printf("Can't find protocol for %s; error.\n", proto);
		return(NULL);
	}

	PRINTF("Registering primitive with %s(%s): %s, %s, %s, %s\n",
		proto, pe->s.name, section, field, op, arg);

	e->prid=pe->reg_prim(pe->field_lookup(section, field), op, arg);

	if(e->prid==-1){
		free(e);
		return(NULL);
	}

	if(parent->kids){
		e->id=(parent->kids->id + 1);
		PRINTF("Setting e->next to %p.\n", parent->kids);
		e->next=parent->kids;
	}
	PRINTF("Setting parent->kids to %p.\n", parent->kids);
	parent->kids=e;

	PRINTF("Registered primitive (section %s, field %s) as %i.\n", section, field, e->prid);

	return(e);
}

/* exported implementations */

static event_id
create_event(void)
{
	event_id id = collections_used++;

	collections[id]=add_element(NULL, ETYPE_COL);
	/* printf("Created event %i.\n", id); */
	return(id | (EH_MATCHER<<24));
}

static handler_id
register_handler(event_id incoming_event, event_id outgoing_event, xmlNodePtr match)
{
	struct melement *n=NULL;

	PRINTF("Registering match rule; target is %i.%i.\n",
		outgoing_event >> 24, outgoing_event & 0xffffff);

	n=collections[(incoming_event & 0xffffff)];
	PRINTF("Building rule under %i...\n", (incoming_event & 0xffffff));
	if(build_rule(n, match, outgoing_event, 0)!=0){
		printf("build_rule() failed!  Exiting.\n");
		exit(-1);
	}

	return(0);
}

static void
handle_event(struct packet *p)
{
	struct melement *t, *u;
	struct event *e, *f, *g;
	unsigned long tot = 0;
	int i = 0;

	/* FIXME: an optimization would be to start from the set of
	 * rules whose protocol prerequisites are met.	delayed until
	 * later.  Sat Jul 21 01:20:24 EDT 2001 */

	for(e=p->q;e;e=e->next){
		if((e->id>>24)==EH_MATCHER){
			PRINTF("Matcher handling event %i.%i.\n",
				(e->id>>24),
				(e->id&0xffffff));
			t=collections[(e->id&0xffffff)];
			if(t->disc==ETYPE_COL){
				PRINTF("Running collection: ");
				for(u=t->kids;u;u=u->next){
					unsigned long es, ef;
					PRINTF("Evaling element %p for packet %p.\n", u, p);
					rdtscl(es);
					if(eval_element(p, u)){
						if(u->event){
							/* this rule matches; raise the event */
							PRINTF("MATCH on %i!  Raising %i.%i.\n",
								e->id&0xffffff, u->event >> 24, u->event & 0xffffff);
							f=p->c->eng->alloc(p->c, sizeof(*f));
							f->id=u->event;
							f->consumed=0;
							f->next=NULL;
							f->payload=NULL;
							for(g=p->q;g->next;g=g->next)
								;
							if(g){
								PRINTF("Setting g->next to %p.\n", f);
								g->next=f;
							} else
								p->q=f;
						} else {
							PRINTF("ERROR: MATCH on %i, but no target to raise.\n", u->event);
						}
					} else {
							PRINTF("NO MATCH on %i!  Not raising %i.%i.\n",
								e->id&0xffffff, u->event >> 24, u->event & 0xffffff);
					}
					rdtscl(ef);
					PRINTF("Eval of element %i took %i cycles (tot %i).\n", i++, ef-es, (tot += (ef-es)));
				}
			} else {
				printf("Aiee!  Unrecognized type at top-level in matcher.  Error.\n");
				/* FIXME: log; error */
			}
			e->consumed=1;
		} else {
			PRINTF("Matcher skipping event %i.%i.\n",
				(e->id>>24),
				(e->id&0xffffff));
		}
	}
	printf("Total for match event: %i cycles.\n", tot);
}

static int
unregister_handler(handler_id handler)
{
	return(-1);
}

static void
decompose(struct packet *p)
{
	struct protocol_engine *pe;
	struct capsule *e;

	if(p->decomposed){
		printf("ERROR: decomposing already-decomposed packet (%i).  Exiting.\n", p->decomposed);
		exit(-1);
	}

	e=p->encaps;
	while(e){
		if((pe=pls.find_pe_num(proto_table, e->payload_protocol))){
			PRINTF("Decomposing for %s.\n", pe->s.name);
			pe->decompose(p);
		} else {
			PRINTF("Finished decomposing (couldn't find %i.%i)\n", 
			e->payload_protocol>>24,
			e->payload_protocol&0xffffff);
			break;
		}
		e=e->next;
	}
	p->decomposed=1;
}

static int
set_handler_target(event_id incoming, handler_id handler, event_id target)
{
	struct melement *n;
	unsigned int i;

	PRINTF("Setting target for %i.%i to %i.%i.\n",
		incoming, handler, target>>24, target&0xffffff);

	n=collections[(incoming & 0xffffff)];

	for(i=0;i<handler;i++)
		n=n->next;

	n->event=target;

	return(0);
}

