/*******************************************************************
 * paengine_netfilter.c: netfilter-based packet acquisition engine *
 * (c) 2001, Todd Lewis, <tlewis@mindspring.com>                   *
 * (c) 2000, Todd Lewis and SecureWorks, www.secureworks.net       *
 * Released under GNU GPL v2.                                      *
 *******************************************************************/

#include "hank.h"
#include "protocols.h"
#include "packet.h"
#include "paengine.h"
#include <libipq/libipq.h>
#include <linux/netfilter.h>
#include <stdlib.h>		/* malloc */
#include <stdio.h>

static char rcsid[] = "$Id: paengine_netfilter.c,v 1.11 2001/11/10 01:45:33 tlewis Exp $";

#define PRINTF(x...)

/* definitions */
#define STATIC_MODULE_NAME paengine_netfilter

#define H_PACKET_SIZE 10240
static unsigned char packet_buf[10240];

extern struct memcon *mc;

/* static declarations */

static int my_configure (char *);
static int my_init (void);
static struct packet *my_get_packet (event_id id);
static int my_dispose_packet (struct packet *);
static int my_close (void);
static int my_get_capabilities (void);

static struct ipq_handle *qh;

static struct paengine pae = {
  {
   "paengine",
   "netfilter",
   PAVMAJ,
   PAVMIN,
   &pae},
  my_init,
  my_get_packet,
  my_dispose_packet,
  my_close,
  my_get_capabilities,
  my_configure
};

#ifdef ENABLE_DYN_MODULES
struct smodule *paengine = &(pae.s);
#else /*ENABLE_DYN_MODULES */
struct smodule *STATIC_MODULE_NAME = &(pae.s);
#endif /*ENABLE_DYN_MODULES */

/* external declaration */

/* implementations */

/* No configuration for now. */
static int
my_configure (char *config)
{
  if (config == 0)
    return (-1);
  return (0);
}

static int
my_init (void)
{
  if (!(qh = ipq_create_handle (0)))
    {
      fprintf (stderr, "ERROR: Can't create netlink socket, exiting...\n");
      fprintf (stderr, "(Perhaps this kernel doesn't support netlink?)\n");
      exit (-1);
    }

  if (ipq_set_mode (qh, IPQ_COPY_PACKET, 0xFFFF) < 0)
    {
      fprintf (stderr, "ERROR: Can't set netlink mode, exiting...\n");
      fprintf (stderr, "(Perhaps this kernel doesn't support netlink?)\n");
      exit (-1);
    }

  return (0);
}

static struct packet *
my_get_packet (event_id id)
{
  struct memcon *n = mc;
  struct packet *p;
  struct capsule *cap;
  struct memcon_context *c;
  unsigned char *buf;
  ipq_packet_msg_t *nhdr;
  int ptype, error;

  PRINTF ("netfilter:get_packet()\n");

  /* create a new memcon */
  c = n->create (0, "pae-pcap");
  if (c == NULL)
    {
      /* FIXME: log */
      fprintf (stderr, "AIEE!  Can't create memcon!\n");
      exit (-1);
    }

  /* create a new packet */
  if ((p = c->eng->alloc (c, sizeof (struct packet))) == NULL)
    {
      /* FIXME: log */
      fprintf (stderr, "AIEE!  Can't allocate packet memory!\n");
      exit (-1);
    }

  /* create a new capsule */
  if ((cap = c->eng->alloc (c, sizeof (struct capsule))) == NULL)
    {
      /* FIXME: log */
      fprintf (stderr, "AIEE!  Can't allocate capsule memory!\n");
      exit (-1);
    }

/*
  if ((buf = c->eng->alloc (c, H_PACKET_SIZE)) == NULL)
    {
      fprintf (stderr, "AIEE!  Can't allocate ipq buffer!  Exiting.\n");
      exit (-1);
    }
*/
	buf=&packet_buf[0];

ipq_read_repeat:
  while((int32_t)(cap->length = ipq_read (qh, buf, H_PACKET_SIZE, 0)) < 0)
	printf("ipq_read() failed; repeating.\n");
	

  if (cap->length == 0)
    {
      fprintf (stderr, "paengine_netfilter: read timeout exceeded.\n");
      return (NULL);
    }
  else
    {
      ptype = ipq_message_type (buf);
      nhdr = ipq_get_packet (buf);

      switch (ptype)
	{
	case 0:
		goto ipq_read_repeat;
	case IPQM_PACKET:
	  p->ts.tv_sec = nhdr->timestamp_sec;
	  p->ts.tv_usec = nhdr->timestamp_usec;
	  cap->length = nhdr->data_len;
	  cap->payload_length = nhdr->data_len;
	  cap->payload = &(nhdr->payload[0]);
	  PRINTF ("netfilter: payload(1) is at %p.\n", cap->payload);
	  p->id = nhdr->packet_id;
	  break;
	case NLMSG_ERROR:
	  error = -ipq_get_msgerr (buf);
	  fprintf (stderr,
		   "paengine_netfilter->get_packet(): NLMSG_ERROR: %d, %s\n",
		   error, ipq_errstr ());
	  return (NULL);
	default:
	  /* error */
	  printf("Unknown ptype %i; returning NULL.\n", ptype);
	  return (NULL);
	}
    }

  p->paengine = &pae;
  p->c = c;
  p->encaps = cap;
  p->verdict = PA_PACKET_VERDICT_ACCEPT;
  p->decomposed = 0;

  cap->protocol = ((HPN_PAE << 24) | PAE_NETFILTER);
  cap->sections = NULL;
  cap->payload_protocol = ((HPN_PCAP << 24) | DLT_RAW);	/* shd be 12 */

  cap->pe = NULL;
  cap->next = NULL;
  cap->parent = p;
  cap->matches = NULL;

  if ((p->q = c->eng->alloc (c, sizeof (struct event))) == NULL)
    {
      fprintf (stderr, "AIEE!  Can't allocate event queue memory!\n");
      exit (-1);
    }
  p->q->id = id;
  p->q->consumed = 0;
  p->q->payload = NULL;
  p->q->next = NULL;

  PRINTF ("netfilter: payload(2) is at %p.\n", cap->payload);

  return (p);

}

static int
my_dispose_packet (struct packet *p)
{
  int verdict;

  if (p->verdict & PA_PACKET_VERDICT_DROP)
    {
      PRINTF ("Drop!\n");
      verdict = NF_DROP;
    }
  else if (p->verdict & PA_PACKET_VERDICT_REJECT)
    {
      /* FIXME: Call raw socket ICMP reject routine */
      PRINTF ("Reject!\n");
      verdict = NF_DROP;
    }
  else if (p->verdict & PA_PACKET_VERDICT_ACCEPT)
    {
      PRINTF ("Accept.\n");
      verdict = NF_ACCEPT;
    }
  else
    {
      PRINTF ("Unrecognized verdict: %i; drop.\n", p->verdict);
      verdict = NF_DROP;
    }

  PRINTF ("Setting verdict of %i.\n", verdict);
  ipq_set_verdict (qh, p->id, verdict, p->encaps->length, p->encaps->payload);

  p->c->eng->destroy (p->c);
  return (0);
}


static int
my_close (void)
{
  return (ipq_destroy_handle (qh));
}

static int
my_get_capabilities (void)
{
  return (PA_CAP_DUMP | PA_CAP_DISCARD);
}
