/*******************************************************************
 * paengine_pcap.c: pcap-based packet acquisition engine           *
 * (c) 2001, Todd Lewis <tlewis@mindspring.com>                    *
 * Released under GNU GPL v2.                                      *
 *******************************************************************/

/*******************************************************************
 * A new paengine to use pcap, free of all of the snort v1 baggage.*
 *******************************************************************/

#include "hank.h"
#include "configuration.h"
#include "protocols.h"
#include "paengine.h"
#include <string.h> /* strlen */
#include <pcap.h>

#define SIZE_T u_int32_t

static char rcsid[] = "$Id: paengine_pcap.c,v 1.12 2001/11/10 01:45:33 tlewis Exp $";

/* definitions */
#define PCAP_MODE_READFILE 1
#define PCAP_MODE_INTERFACE 2

#define PL_MAX 65535

#define STATIC_MODULE_NAME paengine_pcap

extern struct memcon *mc;

/* data structures */
static struct config {
	int configured;
	int mode;
	char *readfile;
	char *interface;
	pcap_t *pcap;
	u_int32_t dlt;
} config;

static struct {
	struct hank_timeval ts;
	char in_use;
	SIZE_T length;
	unsigned char buf[PL_MAX];
} static_buf;

static int pid=0;

/* declarations */
static int init(void);
static struct packet *get_packet(event_id id);
static int dispose_packet(struct packet *);
static int close(void);
static int get_capabilities(void);
static int configure(char * c);

static struct paengine pae = {
	{
		"paengine",
		"pcap",
		PAVMAJ,
		PAVMIN,
		&pae
	},
	&init,
	&get_packet,
	&dispose_packet,
	&close,
	&get_capabilities,
	&configure
};

#ifdef ENABLE_DYN_MODULES
struct smodule * paengine [2] = {
#else  /*ENABLE_DYN_MODULES*/
struct smodule * STATIC_MODULE_NAME [2] =  {
#endif /*ENABLE_DYN_MODULES*/
	&(pae.s),
	NULL
};

/* private functions */
static void 
split_kv(char *string, char *kv[2])
{
	SIZE_T s, i;

	kv[0]=string;
	kv[1]=0;

	s=strlen(string);
	for(i=0;i<(s-1);i++){
		if(string[i]=='='){
			string[i]=0;
			kv[1]=&(string[i+1]);
			return;
		}
	}
	if(string[s]=='=')
		string[s]=0;

	return;
}

static void
callback(u_char *a, const struct pcap_pkthdr *b, const u_char *c)
{
	static_buf.in_use=1;
	memcpy(&(static_buf.buf), c, b->caplen);
	memcpy(&(static_buf.ts), &(b->ts), sizeof(b->ts));
	static_buf.length=b->caplen;
	a++;
}

/* private functions */
static int 
init(void)
{
	config.configured=0;
	return(0);
}

static struct packet *
get_packet(event_id id)
{
	struct memcon *n=mc;
	struct packet *p;
	struct capsule *cap;
	struct memcon_context *c;
	int i;

	if(static_buf.in_use==1){
		/* FIXME: log and errno */
		fprintf(stderr, "AIEE!  pcap buffer in use; error.\n");
		return(NULL);
	}

	/* do callback */
	if((i=pcap_dispatch(config.pcap, 1, &callback, NULL))==0){
		/* FIXME: log */
		/* fprintf(stderr, "AIEE!  pcap_dispatch returned %i!\n", i); */
		return(NULL);
	}

	/* create a new memcon */
	c=n->create(0, "pae-pcap");
	if(c==NULL){
		/* FIXME: log */
		fprintf(stderr, "AIEE!  Can't create memcon!\n");
		exit(-1);
	}

	/* create a new packet */
	/* Look, Ma!  We're OO! */
	if((p=c->eng->alloc(c, sizeof(struct packet)))==NULL){
		/* FIXME: log */
		fprintf(stderr, "AIEE!  Can't allocate packet memory!\n");
		exit(-1);
	}

	/* create a new capsule */
	if((cap=c->eng->alloc(c, sizeof(struct capsule)))==NULL){
		/* FIXME: log */
		fprintf(stderr, "AIEE!  Can't allocate capsule memory!\n");
		exit(-1);
	}

	p->paengine=&pae;
	p->c=c;
	p->encaps=cap;
	p->verdict=PA_PACKET_VERDICT_NO_ACTION;
	p->id=pid++;
	p->decomposed=0;
	memcpy(&(p->ts), &(static_buf.ts), sizeof(static_buf.ts));

	cap->protocol=(HPN_PAE << 24) | PAE_PCAP;
	cap->sections=NULL;
	cap->payload_protocol=config.dlt;
	cap->length=static_buf.length;
	cap->payload_length=static_buf.length;
	cap->payload=&(static_buf.buf[0]);
	cap->pe=NULL;
	cap->next=NULL;
	cap->parent=p;
	cap->matches=NULL;

	if((p->q=c->eng->alloc(c, sizeof(struct event)))==NULL){
		fprintf(stderr, "AIEE!  Can't allocate event queue memory!\n");
		exit(-1);
	}
	p->q->id=id;
	p->q->consumed=0;
	p->q->payload=NULL;
	p->q->next=NULL;

	return(p);
}

static int 
dispose_packet(struct packet *p)
{
	int status=0;

	static_buf.in_use=0;
	if(p->verdict!=PA_PACKET_VERDICT_NO_ACTION){
		/* That's an error with this pae */
		status=-2;
	}
	/* Man, am I good or what? */
	p->c->eng->destroy(p->c);
	return(status);
}

static int 
close(void)
{
	return(0);
}

static int 
get_capabilities(void)
{
	return(PA_CAP_DUMP);
}

static int 
configure(char * c)
{
	char *kv[2];
	char ebuf[PCAP_ERRBUF_SIZE];

	split_kv(c, kv);

	if(config.configured==1){
		return(-3);
	}

	if(!(strcmp(kv[0], "rf"))){
		config.mode=PCAP_MODE_READFILE;
		config.readfile=strdup(kv[1]);
		if(config.readfile==NULL){
			/* FIXME: log */
			return(-1);
		}
		/* FIXME: open readfile */
		config.pcap=pcap_open_offline(config.readfile, ebuf);
		if(config.pcap==NULL){
			fprintf(stderr, "pcap_open() error: %s\n", ebuf);
			/* FIXME: log */
			return(-1);
		}
		goto good;
	}

	if(!(strcmp(kv[0], "interface"))){
		config.mode=PCAP_MODE_INTERFACE;
		config.interface=strdup(kv[1]);
		if(config.interface==NULL){
			/* FIXME: log */
			return(-1);
		}
		/* FIXME: replace snaplen, promisc and to_ms  with config value */
		config.pcap=pcap_open_live(config.interface, PL_MAX, 1, 0, ebuf);
		if(config.pcap==NULL){
			/* FIXME: log */
			fprintf(stderr, "pcap_open() error: %s\n", ebuf);
			return(-1);
		}
		goto good;
	}

	fprintf(stderr, "Unrecognized config option!  Got %s.\n", c);
	return(-3);

good:
	config.dlt=(HPN_PCAP<<24)|pcap_datalink(config.pcap);
	/* fprintf(stderr, "paengine:configure()- protocol is %o\n", config.dlt); */
	config.configured=1;
	return(0);
}
